// smtprec.cpp

//
// Copyright Sbastien Millet, February 2012.
// See the file "COPYING" for the licence.
//

//
// 1.2: mingw compilation and uses micro-seconds instead of milli-seconds
//      in the log file timestamp and in saved messages file names.
//      Autotooled package.
// 1.1: implemented RSET command
// 1.0: workable program
//

//
// Licence (complete text found in the file "COPYING")
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifdef HAVE_CONFIG_H
#include "../config.h"
#else
#include "../extracfg.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <signal.h>

#ifdef _WIN32
#include <windows.h>
#else
#include <netinet/in.h>
#include <arpa/inet.h>
#endif

#ifndef _MSC_VER
#include <unistd.h>
#include <sys/time.h>
#endif

#include <time.h>
#include <ctype.h>

#ifndef SOCKET_ERROR
#define SOCKET_ERROR -1
#endif

#ifdef _WIN32
typedef int socklen_t;
#endif

const int ERR_STR_BUFSIZE = 200;

void log(const bool& enforce, const char *fmt, ...);


//
// * **************************** *
// * CONSTANTS, DEFINES and TYPES *
// * **************************** *
//

const char *LOGFILE_NAME = PACKAGE_TARNAME ".log";
const int DEFAULT_LISTENING_PORT = 25;
const bool DEFAULT_PRINT_TRACE = true;

  // Internet standard newline
const char *INET_NEW_LINE = "\015\012";
  // Maximum size of an input line in the TCP connection
const int MAX_READLINE_SIZE = 10000;
  // Maximum number of recipients for an email
const int MAX_LIST_RCPT_TO = 30000;

  // Used to build the file name, to save a received email
const char *FILENAME_PREFIX = "Msg-";
const char *FILENAME_POSTFIX = ".txt";

  // Messages replied during SMTP conversation
const char *SMTP_GREETING = "220 SMTP Service ready (" PACKAGE_STRING ")";
const char *SMTP_HELO_NEEDS_ARGUMENT_1 = "501 Command \"%s\" requires an argument";
const char *SMTP_PLEASED_2 = "250 " PACKAGE " Hello %s ([%s]), pleased to meet you";
const char *SMTP_IMPLICIT_HELO = "250-HELO/EHLO Command not issued, proceeding anyway";
const char *SMTP_RSET_RESP = "250 Reset state";
const char *SMTP_NEED_MAILFROM = "503 Issue MAIL FROM: command before RCPT TO: command";
const char *SMTP_MAILFROM_NEEDS_ARGUMENT = "501 Syntax error, parameters in command \"mail from:\" unrecognized or missing";
const char *SMTP_SENDER_OK_1 = "250 %s... Sender OK";
const char *SMTP_TOOMANY_RCPTTO = "501 Not enough room to store recipients";
const char *SMTP_RCPTTO_NEEDS_ARGUMENT = "501 Syntax error, parameters in command \"rcpt to:\" unrecognized or missing";
const char *SMTP_RECIPIENT_OK_1 = "250 %s... Recipient OK";
const char *SMTP_NEED_RCPTTO_BEFORE_DATA = "503 Issue RCPT TO: command before DATA command";
const char *SMTP_ENTER_MESSAGE = "354 Enter message, end with \".\" on a line by itself";
const char *SMTP_ACCEPTED = "250 Message accepted for delivery";
const char *SMTP_CLOSE = "221 " PACKAGE " SMTP Service closing transmission channel";

enum return_code_t {SUCCESS, FAILURE};


//
// * **************** *
// * GLOBAL VARIABLES *
// * **************** *
//

FILE *fLog;
  // Not modifiable at run time at the moment, left as a non-constant
  // to have it modifiable with a parameter, in a future release.
bool print_trace = DEFAULT_PRINT_TRACE;
unsigned int listening_port = DEFAULT_LISTENING_PORT;


//
// * ****************************** *
// * FUNCTIONS, PORTABILITY PURPOSE *
// * ****************************** *
//

#ifdef _MSC_VER
int snprintf(char *buffer, int buf_size, const char *format, ...) {
  va_list args;
  va_start(args, format);

  int res = vsprintf_s(buffer, buf_size, format, args);

  va_end(args);

  return res;
}
#define strcasecmp _stricmp
#define strncasecmp _strnicmp
#endif

//
// Get current micro-second time
//

#ifdef _MSC_VER

const __int64 DELTA_EPOCH_IN_MICROSECS = 11644473600000000;
struct timezone {
  int  tz_minuteswest; /* minutes W of Greenwich */
  int  tz_dsttime;     /* type of dst correction */
};

int MY_GETTIMEOFDAY(struct timeval *tv, struct timezone *tz) {
  FILETIME ft;
  unsigned __int64 tmpres = 0;
  static int tzflag;

  if (NULL != tv) {
    GetSystemTimeAsFileTime(&ft);

    tmpres |= ft.dwHighDateTime;
    tmpres <<= 32;
    tmpres |= ft.dwLowDateTime;

    /*converting file time to unix epoch*/
    tmpres -= DELTA_EPOCH_IN_MICROSECS;
    tmpres /= 10;  /*convert into microseconds*/
    tv->tv_sec = (long)(tmpres / 1000000UL);
    tv->tv_usec = (long)(tmpres % 1000000UL);
  }

  if (NULL != tz) {
    if (!tzflag) {
      _tzset();
      tzflag++;
    }
    tz->tz_minuteswest = _timezone / 60;
    tz->tz_dsttime = _daylight;
  }

  return 0;
}

#define EXIT_FAILURE 1
#define EXIT_SUCCESS 0

#else

#define MY_GETTIMEOFDAY(a, b) gettimeofday(a, b)

#endif

#ifdef _WIN32

//
// Initializes Windows Winsock system
//
enum return_code_t init_network() {
  WORD ver_requested;
  WSADATA wsa_data;

  ver_requested = MAKEWORD(1, 1);
  int e = WSAStartup(ver_requested, &wsa_data);
  if (e != 0) {
    fprintf(stderr, "Error calling WSAStartup()\n");
    return FAILURE;
  }
  if (LOBYTE(wsa_data.wVersion) != 1 || HIBYTE(wsa_data.wVersion) != 1) {
    WSACleanup();
    fprintf(stderr, "Unable to find proper version of Winsock.dll\n");
    return FAILURE;
  }
  return SUCCESS;
}

int os_last_err() {
  int r = WSAGetLastError();
  //WSACleanup();
  return r;
}

char *os_last_err_desc(char *s, size_t s_bufsize) {
  LPVOID lpMsgBuf;
  DWORD last_err = WSAGetLastError();
  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
    NULL, last_err, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR)&lpMsgBuf, 0, NULL);
  char tmp[ERR_STR_BUFSIZE];
  strncpy(tmp, (char *)lpMsgBuf, sizeof(tmp));
  int n = strlen(tmp);
  if (n >= 2) {
    if (tmp[n - 2] == '\r' && tmp[n - 1] == '\n')
      tmp[n - 2] = '\0';
  }
  snprintf(s, s_bufsize, "code=%lu (%s)", last_err, tmp);
  //WSACleanup();
  return s;
}

void os_closesocket(int sock) {
  closesocket(sock);
}

//
// CTRL-C handler (Windows)
//
BOOL WINAPI windows_trap_ctrl_handler(DWORD dwCtrlType) {
  bool quit_app = false;
  switch(dwCtrlType) {
    case CTRL_C_EVENT:
      quit_app = true;
      break;
    case CTRL_CLOSE_EVENT:
      quit_app = true;
      break;
    case CTRL_BREAK_EVENT:
      quit_app = true;
      break;
    default:
      return FALSE;
  }
  if (quit_app) {
    log(true, "(Received termination signal, quitting)");
    exit(-1000);
  }
    // This code is not reachable...
  return FALSE;
}

//
// Return value 0 = ok, non-0 = error
//
int set_termination_handlers() {
  return SetConsoleCtrlHandler(windows_trap_ctrl_handler, true);
}

#else

enum return_code_t init_network() {
  return SUCCESS;
}

int os_last_err() {
  return errno;
}

char *os_last_err_desc(char *s, size_t s_bufsize) {
  snprintf(s, s_bufsize, "code=%i (%s)", errno, strerror(errno));
  return s;
}

void os_closesocket(int sock) {
  close(sock);
}

void sig_handler(int) {
  log(true, "(Received termination signal, quitting)");
  exit(EXIT_FAILURE);
}

int set_termination_handlers() {
  signal(SIGTERM, sig_handler);
  signal(SIGABRT, sig_handler);
  signal(SIGINT, sig_handler);
  return 1;
}

#endif


//
// * ************************** *
// * FUNCTIONS, GENERAL PURPOSE *
// * ************************** *
//

//
// Returns the usec of current time
//
long int get_current_usec() {
  struct timeval tv;
  struct timezone tz;
  if (MY_GETTIMEOFDAY(&tv, &tz) == -1) {
    fprintf(stderr, "get_current_usec(): error executing gettimeofday()!!!\n");
    exit(EXIT_FAILURE);
  }
  return tv.tv_usec;
}

//
// Free (call to free()) if the pointer is not NULL, and set it to NULL
// once freed.
//
void free_if_not_NULL(char* *p) {
  if ((*p) != NULL) {
    free(*p);
    *p = NULL;
  }
}

//
// Return true if s begins with prefix, false otherwise
// String comparison is case insensitive
//
bool s_begins_with(const char *s, const char *begins_with) {
  return (strncasecmp(s, begins_with, strlen(begins_with)) == 0);
}

//
// Removes leading and traling spaces in a *MODIFYABLE* string.
// To remove trailing spaces, it just writes '\0' after the last non-space
// found from right to left.
//
char *trim(char *str) {
  char *end;

  while(isspace(*str)) str++;

  if(*str == 0) {
    return str;
  }

  end = str + strlen(str) - 1;
  while (end > str && isspace(*end))
    end--;

  *(end + 1) = '\0';
  return str;
}

//
// Wait for the millitm part of the timer to change
// => it'll guarantee uniqueness of message file names.
//
void wait_for_usec_to_change() {
  long int u0 = get_current_usec();
  while (u0 == get_current_usec())
    ;
}

//
// Return a filename, created in such a way that it'll always be unique.
//
// FIXME
//
// The uniqueness of names produced is guaranteed by the time with a
// granularity of one millisecond. It is too long to ensure uniqueness
// in case the function is called several times in a short period.
// That's why the function wait_for_ftime_ms_to_change() has been created.
// All this is uggly, yes, yes.
// Well, this code is to debug SMTP, it is not a production SMTP server!
//
// To improve it, the function below should remember with a static variable
// the last time used to produce a file name, and use a counter (static
// variable, again), of two or three digits, to distinguish between calls
// done during the same millisecond. And this secondary counter should
// always be appended to the numeric sequence of the file name.
// Of course when using such a method, and using an integer value for the
// counter, a granularity of one second for the time becomes appropriate.
// Well, well, well.
//
// Also, if the program becomes multi-threaded, there'll be even more than
// this! The thread ID will have to be used at some point.
// Ultimately, a library producing UNIDs should be used, to bring a final
// point to it. That's the right thing to do.
//
// Another (bad) idea would be to use the message unique-ID (use a temp name
// and rename once the message ID is known), that's indeed guaranteed to be
// unique as per RFCs. But this code is to debug SMTP and there is no such
// guarantee for received emails. It is even possible and usual, if the email
// is sent manually through a telnet session, that a received email contains
// no message ID at all.
//
// That's all, falks!
//
char *get_unique_filename_alloc(const char *prefix, const char *postfix) {
  const unsigned int FILENAME_BUFFER_SIZE = 100;
  char *filename_buffer;

  time_t ltime = time(NULL);
  struct tm ts;
  ts = *localtime(&ltime);

  filename_buffer = (char *)malloc(FILENAME_BUFFER_SIZE);
  int long_year = ts.tm_year + 1900;
  snprintf(filename_buffer, FILENAME_BUFFER_SIZE, "%s%04d%02d%02d%02d%02d%02d%06li%s", prefix,
      long_year, ts.tm_mon + 1, ts.tm_mday,
      ts.tm_hour, ts.tm_min, ts.tm_sec, get_current_usec(),
      postfix);
  return filename_buffer;
}


//
// * *************************** *
// * FUNCTIONS, PACKAGE SPECIFIC *
// * *************************** *
//

//
// Output a line in the debug system, whatever it is
//
void log(const bool& enforce, const char *fmt, ...) {
  va_list args;
  va_start(args, fmt);

  time_t ltime = time(NULL);
  struct tm ts;
  ts = *localtime(&ltime);

  int short_year = ts.tm_year % 100;

  int l = strlen(fmt) * 2 + 100;
  char *formatted_string = (char *)malloc(l);
  vsnprintf(formatted_string, l, fmt, args);
  char prefix[50];
  snprintf(prefix, 50, "%02d/%02d/%02d %02d:%02d:%02d.%06li  ",
      ts.tm_mday, ts.tm_mon + 1, short_year,
      ts.tm_hour, ts.tm_min, ts.tm_sec, get_current_usec());

  int l2 = strlen(prefix) + strlen(formatted_string) + 2;
  char *final = (char *)malloc(l2);
  strncpy(final, prefix, l2);
  strncat(final, formatted_string, l2);
  strncat(final, "\n", l2);
  if (enforce || print_trace) {
    fputs(final, stdout);
  }
  if (fLog != NULL) {
    fputs(final, fLog);
      // The goal of this program is to produce as much debug info as
      // possible, including, the cases when it would crash.
    fflush(fLog);
  }

  free(final);
  free(formatted_string);

  va_end(args);
}

//
// Initializes the LOG system
//
void log_open() {
  fLog = fopen(LOGFILE_NAME, "a");
//  log(true, "Log open");
}

//
// Closes the LOG system
//
void log_close() {
//  log(true, "Log close");
  fclose(fLog);
}

//
// Start server listening
//
int server_listen() {
  struct sockaddr_in listen_sa;

  char s_err[ERR_STR_BUFSIZE];

  int sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (sock == -1) {
    fprintf(stderr, "Error creating socket, error %s\n", os_last_err_desc(s_err, sizeof(s_err)));
    return 0;
  }

  int bOptVal = true;
  socklen_t bOptLen = sizeof(int);
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&bOptVal, bOptLen) == SOCKET_ERROR) {
    fprintf(stderr, "Cannot set socket option, error %s\n", os_last_err_desc(s_err, sizeof(s_err)));
    os_closesocket(sock);
    return 0;
  }

  listen_sa.sin_addr.s_addr = htonl(INADDR_ANY);
  listen_sa.sin_port = htons((u_short)listening_port);
  listen_sa.sin_family = AF_INET;

  if (bind(sock, (struct sockaddr*)&listen_sa, sizeof(listen_sa)) == SOCKET_ERROR) {
    fprintf(stderr, "Cannot bind, error %s\n", os_last_err_desc(s_err, sizeof(s_err)));
    os_closesocket(sock);
    return 0;
  }
  if (listen(sock, SOMAXCONN) == SOCKET_ERROR) {
    fprintf(stderr, "Cannot listen, error %s\n", os_last_err_desc(s_err, sizeof(s_err)));
    os_closesocket(sock);
    return 0;
  }
  return sock;
}

//
// Have server accept incoming connections
//
int server_accept(int listen_sock, struct sockaddr_in* remote_sin) {
  socklen_t remote_sin_len;

  char s_err[ERR_STR_BUFSIZE];

  log(true, "(Listening on port %u...)", listening_port);
  remote_sin_len = sizeof(*remote_sin);
  int sock = accept(listen_sock, (struct sockaddr *)remote_sin, &remote_sin_len);
  if (sock == -1) {
    fprintf(stderr, "Cannot accept, error %s\n", os_last_err_desc(s_err, sizeof(s_err)));
    os_closesocket(listen_sock);
    return 0;
  }
  log(true, "(Connection accepted from %s)", inet_ntoa(remote_sin->sin_addr));
  return sock;
}

//
// Receives a line from a socket (terminated by \015\010)
// Return -1 if an error occured, 1 if reading is successful,
// 0 if transmission is closed.
//
int socket_read_line_alloc(int sock, char** out, const bool& trace) {
  const int INITIAL_READLINE_BUFFER_SIZE = 100;

  int size = INITIAL_READLINE_BUFFER_SIZE;
  int i = 0;
  bool cr = false;
  char ch;
  int nb;

  *out = (char *)malloc(sizeof(char) * size);

  for (;;) {
    if ((nb = recv(sock, &ch, 1, 0)) == SOCKET_ERROR) {
      char s_err[ERR_STR_BUFSIZE];
      fprintf(stderr, "Error reading socket, error %s\n", os_last_err_desc(s_err, sizeof(s_err)));
      os_closesocket(sock);
      return -1;
    }

    if (i >= size) {
      if (size * 2 <= MAX_READLINE_SIZE) {
        size *= 2;
        *out = (char *)realloc(*out, size);
      } else {
        (*out)[size - 1] = '\0';
        break;
      }
    }

    if (nb == 0) {
      (*out)[i] = '\0';
      break;
    }

    if (ch == '\n') {
      if (cr && i > 0) {
        i--;
      }
      (*out)[i] = '\0';
      break;
    } else {
      cr = (ch == '\r' ? true : false);
      (*out)[i] = ch;
    }
    i++;
  }


  if (nb == 0) {
    return 0;
  } else {
    if (trace) {
      log(false, "<<< %s", *out);
    }
    return 1;
  }
}

//
// Send a line to a socket
// Return true if it's OK, false otherwise.
// Manage displaying an error code and closing socket.
//
bool line_send(int s, const char *text, int text_len, int flags, const bool& trace) {
  char *tmp;
  int inet_new_line_len = strlen(INET_NEW_LINE);

  int l = text_len + inet_new_line_len + 1;
  tmp = (char *)malloc(l);
  strncpy(tmp, text, l);
  strncat(tmp, INET_NEW_LINE, l);
  text_len += inet_new_line_len;

  if (trace)
    log(false, ">>> %s", text);

  int e = send(s, tmp, text_len, flags);
  free(tmp);
  if (e == SOCKET_ERROR) {
    char s_err[ERR_STR_BUFSIZE];
    fprintf(stderr, "Error sending to socket, error %s\n", os_last_err_desc(s_err, sizeof(s_err)));
    os_closesocket(s);
    return false;
  }
  return true;
}

//
// Speak SMTP
//
void manage_smtp_transaction(int sock, const struct sockaddr_in& remote_sin) {
  enum smtp_status_t {S_WAITING_HELO, S_AFTER_HELO, S_ENTERDATA, S_DATA, S_QUIT};
  const int INITIAL_LIST_RCPT_TO = 1;

  char *response;
  bool extended_smtp;

  char *mail_from = NULL;

  int size_list_rcpt_to = INITIAL_LIST_RCPT_TO;
  char **rcpt_to;
  int i_rcpt_to = 0;

  const size_t SEND_BUFFER_SIZE = 512;
  char send_buffer[SEND_BUFFER_SIZE];
  char *filename = NULL;
  char *arg;

  FILE *save_msg = NULL;

  rcpt_to = (char **)malloc(sizeof(char*) * size_list_rcpt_to);

  if (!line_send(sock, SMTP_GREETING, strlen(SMTP_GREETING), 0, print_trace)) {
    return;
  }

  unsigned long int message_lines_count = 0;
  int read_res = 0;
  enum smtp_status_t status = S_WAITING_HELO;
  while (status != S_QUIT) {
    if ((read_res = socket_read_line_alloc(sock, &response, print_trace && status != S_DATA)) < 0) {
      os_closesocket(sock);
      break;
    }

      // EMAIL IS BEING RECEIVED
    if (status == S_DATA) {
      send_buffer[0] = '\0';

        // The terminaing single dot line belongs to the message, therefore
        // the comparison between response and "." occurs after writing the
        // line in the file.
      if (save_msg != NULL) {
        int l = strlen(response) + 2;
        char *tmp = (char *)malloc(l);
        strncpy(tmp, response, l);
        strncat(tmp, "\n", l);
        size_t sz = strlen(tmp);
        if (fwrite(tmp, sz, 1, save_msg) != 1) {
          log(true, "Error writing message line to record file");
        }
        free(tmp);
        message_lines_count++;
      }

      if (strcmp(response, ".") == 0) {
        fclose(save_msg);
        log(true, "(Message saved, received %u lines)", message_lines_count);

        strncpy(send_buffer, SMTP_ACCEPTED, SEND_BUFFER_SIZE);
        status = S_AFTER_HELO;

        free_if_not_NULL(&mail_from);
        for (int i = 0; i < i_rcpt_to; i++)
          free(rcpt_to[i]);
        i_rcpt_to = 0;

          //
          // FIXME
          //
          // To be robust, this uggly code should be called by the guilty function:
          // get_unique_filename_alloc(). Placing this call here intimately ties
          // manage_smtp_transaction() with the way get_unique_filename_alloc()
          // function works, all this is not ideal.
          // But I prefer to leave it here. I do not like the idea of a minimalistic
          // function (get_unique_filename_alloc()) to take a HUGE delay before
          // returning.
          //
        wait_for_usec_to_change();

      }

      // HELO or EHLO COMMAND
    } else if (strcasecmp(response, "helo") == 0 || strcasecmp(response, "ehlo") == 0) {
      snprintf(send_buffer, SEND_BUFFER_SIZE, SMTP_HELO_NEEDS_ARGUMENT_1, response);
    } else if (s_begins_with(response, "helo ") || s_begins_with(response, "ehlo ")) {
      extended_smtp = s_begins_with(response, "ehlo ");
      arg = trim(response + 5);
      if (strlen(arg) == 0) {
        response[4] = '\0';
        snprintf(send_buffer, SEND_BUFFER_SIZE, SMTP_HELO_NEEDS_ARGUMENT_1, response);
      } else {
        status = S_AFTER_HELO;
        snprintf(send_buffer, SEND_BUFFER_SIZE, SMTP_PLEASED_2, arg, inet_ntoa(remote_sin.sin_addr));
      }

      // MAIL FROM COMMAND
    } else if (s_begins_with(response, "mail from:")) {
      free_if_not_NULL(&mail_from);
      if (status == S_WAITING_HELO) {
        if (!line_send(sock, SMTP_IMPLICIT_HELO, strlen(SMTP_IMPLICIT_HELO), 0, print_trace)) {
          break;
        }
      }
      status = S_AFTER_HELO;
      arg = trim(response + 10);
      if (strlen(arg) == 0) {
        strncpy(send_buffer, SMTP_MAILFROM_NEEDS_ARGUMENT, SEND_BUFFER_SIZE);
      } else {
        int l = strlen(arg) + 1;
        mail_from = (char *)malloc(l);
        strncpy(mail_from, arg, l);
        snprintf(send_buffer, SEND_BUFFER_SIZE, SMTP_SENDER_OK_1, mail_from);
      }

      // RCPT TO COMMAND
    } else if (s_begins_with(response, "rcpt to:")) {
      if (mail_from == NULL) {
        strncpy(send_buffer, SMTP_NEED_MAILFROM, SEND_BUFFER_SIZE);
      } else {
        arg = trim(response + 8);
        if (strlen(arg) == 0) {
          strncpy(send_buffer, SMTP_RCPTTO_NEEDS_ARGUMENT, SEND_BUFFER_SIZE);
        } else {
          if (i_rcpt_to >= size_list_rcpt_to) {
            if (size_list_rcpt_to * 2 <= MAX_LIST_RCPT_TO) {
              size_list_rcpt_to *= 2;
              rcpt_to = (char **)realloc(rcpt_to, sizeof(char*) * size_list_rcpt_to);
            } else {
              strncpy(send_buffer, SMTP_TOOMANY_RCPTTO, SEND_BUFFER_SIZE);
            }
          }
          if (i_rcpt_to < size_list_rcpt_to) {
            int l = strlen(arg) + 1;
            rcpt_to[i_rcpt_to] = (char *)malloc(l);
            strncpy(rcpt_to[i_rcpt_to], arg, l);
            snprintf(send_buffer, SEND_BUFFER_SIZE, SMTP_RECIPIENT_OK_1, rcpt_to[i_rcpt_to]);
            i_rcpt_to++;
          }
        }
      }

      // QUIT COMMAND
    } else if (strcasecmp(response, "quit") == 0) {
      strncpy(send_buffer, SMTP_CLOSE, SEND_BUFFER_SIZE);
      status = S_QUIT;

    } else if (strcasecmp(response, "rset") == 0) {
      strncpy(send_buffer, SMTP_RSET_RESP, SEND_BUFFER_SIZE);
      status = S_AFTER_HELO;

      free_if_not_NULL(&mail_from);
      for (int i = 0; i < i_rcpt_to; i++)
        free(rcpt_to[i]);
      i_rcpt_to = 0;

      // DATA COMMAND
    } else if (strcasecmp(response, "data") == 0) {
      if (i_rcpt_to <= 0) {
        strncpy(send_buffer, SMTP_NEED_RCPTTO_BEFORE_DATA, SEND_BUFFER_SIZE);
      } else {
        strncpy(send_buffer, SMTP_ENTER_MESSAGE, SEND_BUFFER_SIZE);
        status = S_ENTERDATA;
      }

      // EMPTY COMMAND
    } else if (strlen(response) == 0) {
      send_buffer[0] = '\0';

      // ANY OTHER COMMAND (=> syntax error)
    } else {
      snprintf(send_buffer, SEND_BUFFER_SIZE, "500 Syntax error, command \"%s\" unrecognized", response);
    }

    free(response);

    if (send_buffer[0] != '\0') {
      if (!line_send(sock, send_buffer, strlen(send_buffer), 0, print_trace)) {
        break;
      }
    }

    if (status == S_ENTERDATA) {
      free_if_not_NULL(&filename);
      filename = get_unique_filename_alloc(FILENAME_PREFIX, FILENAME_POSTFIX);
      if ((save_msg = fopen(filename, "w")) == NULL) {
        log(true, "Unable to open file %s in 'w' mode", filename);
      } else {
        log(true, "(Will record message in %s)", filename);
        log(false, "...");
        status = S_DATA;
      }
      message_lines_count = 0;
    }

    if (read_res == 0) {
      break;
    }
  }

  free_if_not_NULL(&filename);

  free_if_not_NULL(&mail_from);
  for (int i = 0; i < i_rcpt_to; i++)
    free(rcpt_to[i]);
  free(rcpt_to);

  if (read_res >= 0) {
    os_closesocket(sock);
  }

  log(true, "(Connection closed)");
}

//
// Executed when the program stops
//
void exit_handler() {
  log(true, PACKAGE " stop");
  log_close();
}

//
// Prints help information
//
int print_help() {
  printf(PACKAGE_STRING "\n");
  printf("Copyright Sebastien Millet, February 2012\n");
  printf("\n");
  printf("Usage: " PACKAGE_NAME " [-p port] [-h|/?]\n");
  printf("  Accept incoming connections on SMTP port (or another port if -p is used)\n");
  printf("  and manage SMTP connection with the connecting client.\n");
  printf("  Log all SMTP conversation in %s and save received emails, if any,\n", LOGFILE_NAME);
  printf("  in files of name Msg-<n>.txt, where <n> is a big integer.\n");
  printf("\n");
  printf("      * WARNING *\n");
  printf("  * " PACKAGE " DOES NOT RELAY EMAILS. IT SAVES IT AS THEY ARE RECEIVED *\n");
  printf("  * IT IS MEANT FOR SMTP DEBUGGING PURPOSE ONLY *\n");
  printf("\n");
  printf("This program is free software: you can redistribute it and/or modify\n");
  printf("it under the terms of the GNU General Public License as published by\n");
  printf("the Free Software Foundation, either version 3 of the License, or\n");
  printf("(at your option) any later version.\n");
  printf("\n");
  printf("This program is distributed in the hope that it will be useful,\n");
  printf("but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
  printf("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
  printf("GNU General Public License for more details.\n");
  printf("\n");
  printf("You should have received a copy of the GNU General Public License\n");
  printf("along with this program.  If not, see <http://www.gnu.org/licenses/>.\n");
  printf("\n");
  return 0;
}

//
// Display version information
//
int print_version() {
  printf(PACKAGE_STRING "\n");
  printf("Copyright 2012 Sbastien Millet\n");
	printf("This program is free software; you may redistribute it under the terms of\n");
	printf("the GNU General Public License version 3 or (at your option) any later version.\n");
	printf("This program has absolutely no warranty.\n");
  return 0;
}

//
// Display a small help
//
int usage(const char *argv0) {
  fprintf(stderr, "Unknown parameter(s). Execute %s -h to get help\n", argv0);
  return -1;
}

//
// Entry point
//
int main(int argc, char **argv) {

    // Manage program args
  if (argc == 2)
    if (strcmp(argv[1], "-h") == 0 || strcmp(argv[1], "-?") == 0 || strcmp(argv[1], "/?") == 0
        || strcmp(argv[1], "--help") == 0)
      return print_help();
    else if (strcmp(argv[1], "-v") == 0 || strcmp(argv[1], "--version") == 0)
      return print_version();
    else
      return usage(argv[0]);
  else if (argc == 3)
    if (strcmp(argv[1], "-p") == 0) {
      listening_port = atoi(argv[2]);
      if (listening_port <= 0 || listening_port >= 65535)
        return usage(argv[0]);
    } else
      return usage(argv[0]);
  else if (argc != 1)
    return usage(argv[0]);

    // Manage CTRL-C (or the like) pressing
  if ((set_termination_handlers()) == 0) {
    fprintf(stderr, "Unable to set termination handlers, stopping now\n");
    return -10;
  }

  log_open();
  log(true, PACKAGE_STRING " start");

  if (init_network() != SUCCESS) {
    log_close();
    return -20;
  }
  int listen_sock = server_listen();
  if (listen_sock == 0) {
    log_close();
    return -30;
  }

  atexit(exit_handler);
  struct sockaddr_in remote_sin;
  int sock;
  while (true) {
    sock = server_accept(listen_sock, &remote_sin);
    if (sock != 0)
      manage_smtp_transaction(sock, remote_sin);
  }
  return 0;
}

